using System.Runtime.Intrinsics.Arm;
using System.Security.Cryptography;
using GraphicsManager.Structs;
using OpenTK.Windowing.Desktop;

namespace GraphicsManager.Objects.Core;

public class TextureManager
{
    public static readonly Dictionary<IGLFWGraphicsContext, TextureManager> TextureManagers = new();

    public readonly Dictionary<byte[], Texture> TextureHashMap = new();

    public static TextureManager GetTextureManager(IGLFWGraphicsContext context)
    {
        if (context is null) throw new ArgumentNullException(nameof(context));
        if (!TextureManagers.ContainsKey(context))
        {
            TextureManager tmp = new(context);
            TextureManagers.Add(context, tmp);
            return tmp;
        }

        return TextureManagers[context];
    }

    public IGLFWGraphicsContext Context { get; private set; }

    private TextureManager(IGLFWGraphicsContext context)
    {
        Context = context;
    }

    public IReadOnlyList<Texture> Textures { get => TextureHashMap.Values.ToList().AsReadOnly(); }

    public void RemoveTexture(Texture Texture)
    {
        foreach(KeyValuePair<byte[], Texture> item in TextureHashMap.Where(kvp => kvp.Value == Texture).ToList())
        {
            TextureHashMap.Remove(item.Key);
        }
    }
    
    public Texture AddTexture(byte[] Texture, bool Rounded = false)
    {
        List<byte> hash2 = SHA256.HashData(Texture).ToList();
        hash2.Insert(0, Convert.ToByte(Rounded));
        byte[] hash = hash2.ToArray();
        if (TextureHashMap.ContainsKey(hash)) return TextureHashMap[hash];
        Context.MakeCurrent();
        TextureHashMap.Add(hash, new(Texture, Rounded));
        return TextureHashMap[hash];
    }
}